from __future__ import annotations

import typer

from sqlcheck.cli.commands.parse import parse
from sqlcheck.cli.commands.plan import plan
from sqlcheck.cli.commands.run import run

app = typer.Typer(help="Run SQL test files.", add_completion=False)

app.command()(run)
app.command()(parse)
app.command()(plan)



from __future__ import annotations

from typing import Any

from cel import evaluate

from sqlcheck.function_context import current_context
from sqlcheck.models import FunctionResult


def assess(
    *_args: Any,
    match: str | None = None,
    check: str | None = None,
    **_kwargs: Any,
) -> FunctionResult:
    expression, error_message = _resolve_match_expression(match, check)
    if expression is None:
        return FunctionResult(
            name="assess",
            success=False,
            message=error_message or "Match expression is required for assess()",
        )
    if not isinstance(expression, str):
        return FunctionResult(
            name="assess",
            success=False,
            message="Match expression must be a string",
        )
    context = current_context()
    try:
        result = evaluate(expression, _build_evaluation_context(context))
    except Exception as exc:  # noqa: BLE001 - surface CEL evaluation errors
        return FunctionResult(
            name="assess",
            success=False,
            message=f"Match expression {expression!r} failed: {exc}",
        )
    if not isinstance(result, bool):
        return FunctionResult(
            name="assess",
            success=False,
            message=f"Match expression {expression!r} did not evaluate to a boolean",
        )
    if not result:
        debug_info = _get_debug_info(expression, _build_evaluation_context(context))
        message = f"Match expression {expression!r} evaluated to false"
        if debug_info:
            message += f"\n  {debug_info}"
        return FunctionResult(
            name="assess",
            success=False,
            message=message,
        )
    return FunctionResult(name="assess", success=True)


def _resolve_match_expression(
    match: str | None,
    check: str | None,
) -> tuple[str | None, str | None]:
    if match and check and match != check:
        return None, "Provide only one of match or check"
    return match or check, None


def _get_debug_info(expression: str, eval_context: dict[str, Any]) -> str:
    """Show values of context variables that appear in the expression."""
    import re

    # Find which context variables are referenced in the expression
    relevant_vars = {}
    for var_name, var_value in eval_context.items():
        # Check if this variable appears in the expression
        if re.search(rf"\b{re.escape(var_name)}\b", expression):
            relevant_vars[var_name] = var_value

    if not relevant_vars:
        return ""

    # Format the output
    parts = []
    for var_name, var_value in relevant_vars.items():
        # Try to find the full expression with this variable (e.g., rows[0][0])
        # by looking for patterns like var_name[...][...]
        pattern = rf"{re.escape(var_name)}(?:\[[^\]]+\])*"
        matches = re.findall(pattern, expression)

        for match in matches:
            try:
                evaluated = evaluate(match, eval_context)
                parts.append(f"{match} = {evaluated!r}")
            except Exception:  # noqa: BLE001, S110
                pass

    return ", ".join(parts) if parts else ""


def _build_evaluation_context(context: Any) -> dict[str, Any]:
    status = context.status
    output = context.output
    sql_parsed = context.sql_parsed
    status_label = "success" if status.success else "fail"
    return {
        "status": status_label,
        "success": status.success,
        "returncode": status.returncode,
        "error_code": str(status.returncode),
        "duration_s": status.duration_s,
        "elapsed_ms": int(status.duration_s * 1000),
        "stdout": output.stdout,
        "stderr": output.stderr,
        "error_message": output.stderr,
        "rows": output.rows,
        "output": {
            "stdout": output.stdout,
            "stderr": output.stderr,
            "rows": output.rows,
        },
        "sql": sql_parsed.source,
        "statements": [statement.text for statement in sql_parsed.statements],
        "statement_count": len(sql_parsed.statements),
    }



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Iterable

from sqlcheck.models import DirectiveCall, SQLParsed, SQLSegment, SQLStatement


class DirectiveParseError(ValueError):
    pass


def _split_statements(sql: str) -> list[SQLStatement]:
    """Split SQL source into individual statements, respecting string literals."""
    statements: list[SQLStatement] = []
    buffer: list[str] = []
    start = 0
    in_single = False
    in_double = False
    escape = False

    for idx, char in enumerate(sql):
        if escape:
            buffer.append(char)
            escape = False
            continue
        if char == "\\":
            escape = True
            buffer.append(char)
            continue
        if char == "'" and not in_double:
            in_single = not in_single
        elif char == '"' and not in_single:
            in_double = not in_double
        if char == ";" and not in_single and not in_double:
            text = "".join(buffer).strip()
            if text:
                statements.append(SQLStatement(len(statements), text, start, idx))
            buffer = []
            start = idx + 1
        else:
            buffer.append(char)

    tail = "".join(buffer).strip()
    if tail:
        statements.append(SQLStatement(len(statements), tail, start, len(sql)))
    return statements


@dataclass(frozen=True)
class ParsedFile:
    sql_parsed: SQLParsed
    directives: list[DirectiveCall]
    segments: list[SQLSegment]
    template_vars: dict[str, str] = field(default_factory=dict)


def parse_file(path: Path, template_vars: dict[str, str] | None = None) -> ParsedFile:
    """
    Parse a SQL test file, extracting directives and SQL statements.

    Uses Jinja2 to render template variables and extract directive function calls.

    Args:
        path: Path to SQL file
        template_vars: Optional template variables for Jinja rendering

    Returns:
        ParsedFile containing directives, SQL, and segments

    Raises:
        DirectiveParseError: If template rendering or parsing fails
    """
    source = path.read_text(encoding="utf-8")

    # Render with Jinja, extracting both variables and directives
    from sqlcheck.template import render_template, TemplateRenderError

    try:
        rendered, directive_markers = render_template(source, template_vars)
    except TemplateRenderError as exc:
        raise DirectiveParseError(
            f"Template rendering failed for {path}: {exc}"
        ) from exc

    # Convert markers to DirectiveCall objects
    directives = [
        DirectiveCall(
            name=marker.name,
            args=marker.args,
            kwargs=marker.kwargs,
            raw=marker.raw_text
        )
        for marker in directive_markers
    ]

    # Check for deprecated config() directive
    if any(directive.name == "config" for directive in directives):
        raise DirectiveParseError(
            "config() is not supported; use exit_on_failure on directives"
        )

    # Strip directive placeholders to get clean SQL
    sql_source = _strip_directive_placeholders(rendered, directive_markers)

    # Parse SQL statements
    statements = _split_statements(sql_source)
    sql_parsed = SQLParsed(source=sql_source, statements=statements)

    # Segment SQL with directives
    segments = _segment_sql_with_markers(rendered, directive_markers, directives)

    return ParsedFile(
        sql_parsed=sql_parsed,
        directives=directives,
        segments=segments,
        template_vars=template_vars or {},
    )


def _strip_directive_placeholders(rendered: str, markers: list) -> str:
    """Remove directive placeholders from rendered SQL."""
    result = rendered
    for marker in markers:
        result = result.replace(marker.placeholder, '')
    return result.strip()


def _segment_sql_with_markers(
    rendered: str,
    markers: list,
    directives: list[DirectiveCall]
) -> list[SQLSegment]:
    """
    Pair SQL blocks with their controlling directives.

    Handles:
    - Directive BEFORE SQL (new style): {{ assess() }} SELECT 1;
    """
    if not markers:
        # No directives, create default segment with all SQL
        sql_source = _strip_directive_placeholders(rendered, [])
        if sql_source.strip():
            statements = _split_statements(sql_source)
            default_directive = DirectiveCall(name="success", args=(), kwargs={}, raw="")
            return [SQLSegment(
                sql_parsed=SQLParsed(source=sql_source, statements=statements),
                directive=default_directive
            )]
        return []

    segments: list[SQLSegment] = []
    chunks: list[str] = []
    cursor = 0

    for marker in markers:
        marker_pos = rendered.find(marker.placeholder, cursor)
        if marker_pos == -1:
            continue
        chunks.append(rendered[cursor:marker_pos])
        cursor = marker_pos + len(marker.placeholder)
    chunks.append(rendered[cursor:])

    for index, directive in enumerate(directives):
        sql_after = chunks[index + 1].strip() if index + 1 < len(chunks) else ""
        if not sql_after:
            continue
        statements = _split_statements(sql_after)
        segments.append(SQLSegment(
            sql_parsed=SQLParsed(source=sql_after, statements=statements),
            directive=directive
        ))

    return segments


def summarize_directives(directives: Iterable[DirectiveCall]) -> dict[str, Any]:
    """Extract metadata (name, tags, serial, timeout, retries) from directives."""
    summary: dict[str, Any] = {
        "serial": False,
        "timeout": None,
        "retries": 0,
        "tags": [],
        "name": None,
    }
    for directive in directives:
        if "serial" in directive.kwargs:
            summary["serial"] = summary["serial"] or bool(directive.kwargs["serial"])
        if "timeout" in directive.kwargs:
            summary["timeout"] = max(summary["timeout"] or 0, float(directive.kwargs["timeout"]))
        if "retries" in directive.kwargs:
            summary["retries"] = max(summary["retries"], int(directive.kwargs["retries"]))
        if "tags" in directive.kwargs:
            tags = directive.kwargs["tags"]
            if isinstance(tags, str):
                summary["tags"].append(tags)
            else:
                summary["tags"].extend(list(tags))
        if "name" in directive.kwargs and not summary["name"]:
            summary["name"] = str(directive.kwargs["name"])
    return summary



from dataclasses import dataclass, field
from typing import Any, Literal, Optional


@dataclass(frozen=True)
class ExecMeta:
    elapsed_ms: int
    rowcount: int | None
    columns: list[str] | None


# ----------------------------
# DBConnection
# ----------------------------

FetchMode = Literal["auto", True, False]



import threading
from typing import Any, Optional

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.pool import NullPool

# Cache key includes URL + poolclass so we never mix pooled vs NullPool engines
_ENGINE_CACHE: dict[tuple[str, type | None], Engine] = {}
_ENGINE_CACHE_LOCK = threading.Lock()


def dispose_all_engines() -> None:
    """
    Force-close pooled connections and clear the engine cache.
    Useful in tests.
    """
    with _ENGINE_CACHE_LOCK:
        for eng in _ENGINE_CACHE.values():
            try:
                eng.dispose()
            except Exception:
                pass
        _ENGINE_CACHE.clear()


def get_engine(url: str, *, engine_kwargs: Optional[dict[str, Any]] = None) -> Engine:
    """
    Return a cached Engine for this URL (thread-safe).

    Default behavior:
    - Reuse engines (and their pools) across DBConnection instances for the same URL.
    - For SQLite in-memory URLs, auto-inject NullPool unless user overrides poolclass,
        so each `with DBConnection(...):` gets a truly fresh in-memory database.
    """
    engine_kwargs = dict(engine_kwargs or {})  # copy so we can safely mutate

    # Auto-disable pooling for in-memory SQLite unless user explicitly sets poolclass.
    if ":memory:" in url and "poolclass" not in engine_kwargs:
        engine_kwargs["poolclass"] = NullPool

    poolclass = engine_kwargs.get("poolclass")  # might be None
    cache_key = (url, poolclass)

    with _ENGINE_CACHE_LOCK:
        engine = _ENGINE_CACHE.get(cache_key)
        if engine is None:
            # pool_pre_ping is useful for long-lived CLIs; sqlite will ignore it.
            engine = create_engine(url, pool_pre_ping=True, **engine_kwargs)
            _ENGINE_CACHE[cache_key] = engine
        return engine



from __future__ import annotations

from pathlib import Path

import typer

from sqlcheck.discovery import build_test_case, discover_files
from sqlcheck.models import TestCase


def discover_cases(target: Path, pattern: str, template_vars: dict[str, str] | None = None) -> list[TestCase]:
    paths = discover_files(target, pattern)
    if not paths:
        print("No test files found.")
        raise typer.Exit(code=1)
    return [build_test_case(path, template_vars=template_vars) for path in paths]


__all__ = ["discover_cases"]



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any


@dataclass(frozen=True)
class SQLStatement:
    index: int
    text: str
    start: int
    end: int


@dataclass(frozen=True)
class SQLParsed:
    source: str
    statements: list[SQLStatement]


@dataclass(frozen=True)
class DirectiveCall:
    name: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    raw: str


@dataclass(frozen=True)
class SQLSegment:
    sql_parsed: SQLParsed
    directive: DirectiveCall


@dataclass(frozen=True)
class TestMetadata:
    name: str
    tags: list[str] = field(default_factory=list)
    serial: bool = False
    timeout: float | None = None
    retries: int = 0
    template_vars: dict[str, str] = field(default_factory=dict)


@dataclass(frozen=True)
class TestCase:
    path: Path
    sql_parsed: SQLParsed
    directives: list[DirectiveCall]
    segments: list[SQLSegment]
    metadata: TestMetadata


@dataclass(frozen=True)
class ExecutionStatus:
    success: bool
    returncode: int
    duration_s: float


@dataclass(frozen=True)
class ExecutionOutput:
    stdout: str
    stderr: str
    rows: list[list[Any]] = field(default_factory=list)


@dataclass(frozen=True)
class ExecutionResult:
    status: ExecutionStatus
    output: ExecutionOutput


@dataclass(frozen=True)
class FunctionResult:
    name: str
    success: bool
    message: str | None = None


@dataclass(frozen=True)
class TestResult:
    case: TestCase
    status: ExecutionStatus
    output: ExecutionOutput
    function_results: list[FunctionResult]

    @property
    def success(self) -> bool:
        return all(result.success for result in self.function_results)



import os
import re
from pathlib import Path
from typing import Any

import yaml
from sqlalchemy.engine import URL

from .constants import DEFAULT_CONN_IDS, LIBRARY_CONNECTIONS, ENV_PREFIXS


def normalize_connection_name(name: str) -> str:
    """
    Normalize connection IDs so env var names are stable:
    - uppercase
    - non-alphanumeric -> underscore
    - collapse underscores
    - trim underscores
    """
    s = name.strip().upper()
    s = re.sub(r"[^A-Z0-9]+", "_", s)
    s = re.sub(r"_+", "_", s).strip("_")
    return s


def _expand_env_vars(value: Any) -> Any:
    """
    Recursively expand environment variables in strings, dicts, and lists.

    Expands ${VAR_NAME} and $VAR_NAME patterns in strings.
    """
    if isinstance(value, str):
        return os.path.expandvars(value)
    elif isinstance(value, dict):
        return {k: _expand_env_vars(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [_expand_env_vars(item) for item in value]
    else:
        return value


def _load_connections_from_yaml(yaml_path: str) -> dict[str, dict[str, Any]]:
    """
    Load connection definitions from YAML file.
    Returns a dict mapping conn_id -> URL.create() parameters.
    Expands environment variables in all string values.
    """
    expanded_path = Path(yaml_path).expanduser()
    if not expanded_path.exists():
        return {}

    try:
        with open(expanded_path, "r") as f:
            data = yaml.safe_load(f)
            if isinstance(data, dict):
                # Expand environment variables in all values
                return _expand_env_vars(data)
            return {}
    except Exception:
        return {}


def _build_not_found_error(conn_id: str, normalized_name: str) -> str:
    """
    Build a comprehensive error message listing all search locations.

    Args:
        conn_id: Original connection ID
        normalized_name: Normalized version used for env var lookups

    Returns:
        Multi-line error message showing where we looked
    """
    env_vars_checked = [f"{prefix}{normalized_name}" for prefix in ENV_PREFIXS]

    message_parts = [
        f"No connection URL found for conn_id={conn_id!r}.",
        "",
        "Searched environment variables:",
    ]
    for env_var in env_vars_checked:
        message_parts.append(f"  - {env_var}")

    message_parts.append("")
    message_parts.append("Searched YAML files:")
    for yaml_file in LIBRARY_CONNECTIONS:
        expanded = Path(yaml_file).expanduser()
        exists = " (found)" if expanded.exists() else " (not found)"
        message_parts.append(f"  - {yaml_file}{exists}")

    return "\n".join(message_parts)


def resolve_connection_url(conn_id_or_url: str | None) -> str:
    """
    Resolve a connection ID to a database URL.

    Resolution order (first match wins):
    1. Use DEFAULT_CONN_IDS if conn_id_or_url is None
    2. Direct URL (if contains "://")
    3. File path (.csv or .xlsx) - returns DuckDB with special marker
    4. Environment variables (all ENV_PREFIXS in order)
    5. YAML files (all LIBRARY_CONNECTIONS in order)

    Args:
        conn_id_or_url: Connection ID to resolve, a direct database URL, file path, or None

    Returns:
        Database URL string (for file paths, returns "duckdb:///:memory:?file_path=<path>")

    Raises:
        ValueError: If connection ID cannot be resolved

    Example:
        # Direct URL passthrough
        url = resolve_connection_url("postgresql://localhost/db")

        # Environment variable lookup
        os.environ["SQLCOMPARE_CONN_PROD"] = "postgresql://prod/db"
        url = resolve_connection_url("prod")

        # Use default connection if None
        os.environ["SQLCOMPARE_CONN_DEFAULT"] = "my_conn"
        url = resolve_connection_url(None)

        # File path (CSV or Excel)
        url = resolve_connection_url("/path/to/data.csv")

        # YAML file lookup (if not in env)
        # Searches ~/.sqlcompare/connections.yml, then ~/.dtk/connections.yml
    """
    # Step 0: Use default connection if None (search all DEFAULT_CONN_IDS)
    if conn_id_or_url is None:
        for default_id in DEFAULT_CONN_IDS:
            if default_id:
                conn_id_or_url = default_id
                break

        if conn_id_or_url is None:
            raise ValueError(
                "No connection ID provided and no default connection found. "
                "Set SQLCOMPARE_CONN_DEFAULT or DTK_CONN_DEFAULT"
            )

    # Step 1: Direct URL passthrough
    if "://" in conn_id_or_url:
        return conn_id_or_url

    # Step 2: Check if it's a file path (.csv or .xlsx)
    file_path = Path(conn_id_or_url).expanduser()
    if file_path.exists() and file_path.suffix.lower() in (".csv", ".xlsx"):
        # Return DuckDB URL with file path marker
        return f"duckdb:///:memory:?_file_path={file_path.absolute()}"

    normalized_name = normalize_connection_name(conn_id_or_url)

    # Step 3: Search ALL environment prefixes (in order)
    for env_prefix in ENV_PREFIXS:
        env_key = f"{env_prefix}{normalized_name}"
        url_str = os.getenv(env_key)
        if url_str:
            return url_str

    # Step 4: Search ALL YAML files (in order)
    for yaml_path in LIBRARY_CONNECTIONS:
        connections = _load_connections_from_yaml(yaml_path)

        # YAML uses raw conn_id (not normalized) for matching
        if conn_id_or_url in connections:
            url_params = connections[conn_id_or_url]
            try:
                url = URL.create(**url_params)
                return str(url)
            except Exception as e:
                raise ValueError(
                    f"Invalid connection parameters for conn_id={conn_id_or_url!r} "
                    f"in {yaml_path}: {e}"
                ) from e

    # Step 5: Not found anywhere - build helpful error message
    raise ValueError(_build_not_found_error(conn_id_or_url, normalized_name))



from __future__ import annotations

import json
from pathlib import Path

import typer

from sqlcheck.cli.discovery import discover_cases
from sqlcheck.reports import build_plan_payload, write_case_plan


def plan(
    target: Path = typer.Argument(..., help="Target file or directory to scan"),
    pattern: str = typer.Option(
        "**/*.sql", help="Glob pattern for test discovery (default: **/*.sql)"
    ),
    plan_dir: Path | None = typer.Option(
        None, "--plan-dir", help="Write per-test plan JSON files to this directory"
    ),
    json_path: Path | None = typer.Option(
        None, "--json", help="Write plan output to path"
    ),
) -> None:
    cases = discover_cases(target, pattern)
    payload = [build_plan_payload(case) for case in cases]

    if plan_dir:
        plan_dir.mkdir(parents=True, exist_ok=True)
        for case in cases:
            relative_name = str(case.path).replace("/", "__").replace("\\", "__")
            plan_path = plan_dir / f"{relative_name}.plan.json"
            write_case_plan(case, plan_path)

    output = json.dumps(payload, indent=2)
    if json_path:
        json_path.write_text(output, encoding="utf-8")
    elif not plan_dir:
        print(output)



from __future__ import annotations

import importlib
from typing import Iterable

from sqlcheck.function_registry import FunctionRegistry


def load_plugins(modules: Iterable[str], registry: FunctionRegistry) -> None:
    for module_name in modules:
        module = importlib.import_module(module_name)
        register = getattr(module, "register", None)
        if callable(register):
            register(registry)
        else:
            raise ValueError(f"Plugin module '{module_name}' has no register(registry) function")



from __future__ import annotations

from typing import Callable

from sqlcheck.functions.assess import assess
from sqlcheck.functions.fail import fail
from sqlcheck.functions.success import success
from sqlcheck.models import FunctionResult

FunctionType = Callable[..., FunctionResult]


class FunctionRegistry:
    def __init__(self) -> None:
        self._functions: dict[str, Callable[..., FunctionResult]] = {}

    def register(self, name: str, func: Callable[..., FunctionResult]) -> None:
        self._functions[name] = func

    def resolve(self, name: str) -> Callable[..., FunctionResult]:
        if name not in self._functions:
            raise KeyError(f"Unknown function '{name}'")
        return self._functions[name]


def default_registry() -> FunctionRegistry:
    registry = FunctionRegistry()
    registry.register("success", success)
    registry.register("fail", fail)
    registry.register("assess", assess)
    return registry



"""Default configuration values for the db module."""

import os
from pathlib import Path

lib_name = "sqlcheck"
lib_config_folder = Path.home() / ".config" / f"{lib_name}"

LIBRARY_CONNECTIONS = [
    os.getenv(
        f"{lib_name.upper()}_CONNECTIONS_FILE", str(lib_config_folder / "connections.yaml")
    ),
    os.getenv("DTK_CONNECTIONS_FILE", str(Path.home() / ".dtk" / "connections.yml")),
]

ENV_PREFIXS = [
    f"{lib_name.upper()}_CONN_",
    "DTK_CONN_",
]

# Default connection name used when conn_id is None
DEFAULT_CONN_IDS = [
    os.getenv(f"{lib_name.upper()}_CONN_DEFAULT"),
    os.getenv("DTK_CONN_DEFAULT"),
]



from __future__ import annotations

from pathlib import Path

from sqlcheck.models import DirectiveCall, TestCase, TestMetadata
from sqlcheck.parser import ParsedFile, parse_file, summarize_directives


def discover_files(target: Path, pattern: str) -> list[Path]:
    if target.is_file():
        return [target]
    return sorted(target.rglob(pattern))


def build_test_case(path: Path, template_vars: dict[str, str] | None = None) -> TestCase:
    parsed: ParsedFile = parse_file(path, template_vars=template_vars)
    directives = parsed.directives or [DirectiveCall(name="success", args=(), kwargs={}, raw="")]
    summary = summarize_directives(directives)
    metadata = TestMetadata(
        name=summary["name"] or path.stem,
        tags=summary["tags"],
        serial=summary["serial"],
        timeout=summary["timeout"],
        retries=summary["retries"],
        template_vars=template_vars or {},
    )
    return TestCase(
        path=path,
        sql_parsed=parsed.sql_parsed,
        directives=directives,
        segments=parsed.segments,
        metadata=metadata,
    )



from __future__ import annotations

from pathlib import Path

import typer

from sqlcheck.cli.discovery import discover_cases
from sqlcheck.cli.output import print_results
from sqlcheck.function_registry import default_registry
from sqlcheck.plugins import load_plugins
from sqlcheck.reports import write_json, write_junit, write_plan
from sqlcheck.runner import run_cases


def _parse_template_vars(vars_list: list[str]) -> dict[str, str]:
    """
    Parse template variables from key=value strings.

    Args:
        vars_list: List of "key=value" strings

    Returns:
        Dictionary mapping variable names to values

    Raises:
        typer.BadParameter: If any variable is malformed
    """
    variables = {}
    for var_str in vars_list:
        if "=" not in var_str:
            raise typer.BadParameter(
                f"Invalid variable format: {var_str!r}. "
                "Expected format: key=value"
            )
        key, _, value = var_str.partition("=")
        key = key.strip()
        value = value.strip()
        if not key:
            raise typer.BadParameter(
                f"Invalid variable format: {var_str!r}. "
                "Variable name cannot be empty"
            )
        if key in variables:
            raise typer.BadParameter(
                f"Duplicate variable: {key!r}"
            )
        variables[key] = value
    return variables


def run(
    target: Path = typer.Argument(..., help="Target file or directory to scan"),
    pattern: str = typer.Option(
        "**/*.sql", help="Glob pattern for test discovery (default: **/*.sql)"
    ),
    workers: int = typer.Option(5, help="Number of worker threads"),
    connection: str = typer.Option(
        ...,
        "--connection",
        "-c",
        help="Connector name for SQLCHECK_CONN_<NAME> environment lookup",
    ),
    json_path: Path | None = typer.Option(
        None, "--json", help="Write JSON report to path"
    ),
    junit_path: Path | None = typer.Option(
        None, "--junit", help="Write JUnit XML report to path"
    ),
    plan_dir: Path | None = typer.Option(
        None, "--plan-dir", help="Write per-test plan JSON files to this directory"
    ),
    plugin: list[str] | None = typer.Option(
        None, "--plugin", help="Plugin module path to load (can be repeated)"
    ),
    vars: list[str] | None = typer.Option(
        None,
        "--vars",
        "-v",
        help="Template variables in key=value format (repeatable)",
    ),
) -> None:
    # Parse template variables
    template_vars = _parse_template_vars(vars or [])

    cases = discover_cases(target, pattern, template_vars=template_vars)

    registry = default_registry()
    if plugin:
        load_plugins(plugin, registry)

    results = run_cases(cases, connection, registry, workers=workers)

    print_results(results, engine=connection)

    if json_path:
        write_json(results, json_path)
    if junit_path:
        write_junit(results, junit_path)
    if plan_dir:
        plan_dir.mkdir(parents=True, exist_ok=True)
        for result in results:
            relative_name = str(result.case.path).replace("/", "__").replace("\\", "__")
            plan_path = plan_dir / f"{relative_name}.plan.json"
            write_plan(result, plan_path)

    failures = [result for result in results if not result.success]
    if failures:
        raise typer.Exit(code=1)



from __future__ import annotations

import json
from pathlib import Path

import typer

from sqlcheck.parser import parse_file
from sqlcheck.runner import discover_files


def parse(
    target: Path = typer.Argument(..., help="Target file or directory to scan"),
    pattern: str = typer.Option(
        "**/*.sql", help="Glob pattern for test discovery (default: **/*.sql)"
    ),
    json_path: Path | None = typer.Option(
        None, "--json", help="Write parse output to path"
    ),
) -> None:
    paths = discover_files(target, pattern)
    if not paths:
        print("No test files found.")
        raise typer.Exit(code=1)

    payload = []
    for path in paths:
        parsed = parse_file(path)
        payload.append(
            {
                "path": str(path),
                "sql_source": parsed.sql_parsed.source,
                "statements": [
                    {
                        "index": stmt.index,
                        "text": stmt.text,
                        "start": stmt.start,
                        "end": stmt.end,
                    }
                    for stmt in parsed.sql_parsed.statements
                ],
                "directives": [
                    {
                        "name": directive.name,
                        "args": directive.args,
                        "kwargs": directive.kwargs,
                        "raw": directive.raw,
                    }
                    for directive in parsed.directives
                ],
            }
        )

    output = json.dumps(payload, indent=2)
    if json_path:
        json_path.write_text(output, encoding="utf-8")
    else:
        print(output)



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Iterator

from sqlcheck.models import ExecutionOutput, ExecutionStatus, SQLParsed

_context: ContextVar["ExecutionContext | None"] = ContextVar(
    "sqlcheck_execution_context",
    default=None,
)


@dataclass(frozen=True)
class ExecutionContext:
    sql_parsed: SQLParsed
    status: ExecutionStatus
    output: ExecutionOutput


@contextmanager
def execution_context(
    sql_parsed: SQLParsed,
    status: ExecutionStatus,
    output: ExecutionOutput,
) -> Iterator[None]:
    token = _context.set(ExecutionContext(sql_parsed=sql_parsed, status=status, output=output))
    try:
        yield
    finally:
        _context.reset(token)


def current_context() -> ExecutionContext:
    context = _context.get()
    if context is None:
        raise RuntimeError("Function execution context is not set")
    return context



from __future__ import annotations

from typing import Any

from sqlcheck.functions.assess import assess
from sqlcheck.models import FunctionResult


def success(
    *_args: Any,
    match: str | None = None,
    **_kwargs: Any,
) -> FunctionResult:
    expression = "status == 'success'"
    if match:
        expression = f"({expression}) && ({match})"
    result = assess(match=expression)
    return FunctionResult(name="success", success=result.success, message=result.message)



from __future__ import annotations

import concurrent.futures
import time
from typing import Iterable

from sqlcheck.db import DBConnection, DBConnectionError
from sqlcheck.function_context import execution_context
from sqlcheck.function_registry import FunctionRegistry
from sqlcheck.models import (
    ExecutionOutput,
    ExecutionResult,
    ExecutionStatus,
    FunctionResult,
    SQLParsed,
    TestCase,
    TestResult,
)


def run_test_case(
    case: TestCase,
    connection_id: str,
    registry: FunctionRegistry,
) -> TestResult:
    if not case.segments:
        # If there are no segments, it means no directives were parsed,
        # typically because SQL was found before the first directive.
        # This is considered a test failure as the intended checks can't run.
        status = ExecutionStatus(success=False, returncode=1, duration_s=0.0)
        output = ExecutionOutput(
            stdout="",
            stderr="There is no valid SQL statement after the check declaration.",
            rows=[],
        )
        # Return a TestResult indicating failure
        return TestResult(
            case=case,
            status=status,
            output=output,
            function_results=[
                FunctionResult(
                    name="parse_failure",
                    success=False,
                    message="There is no valid SQL statement after the check declaration.",
                )
            ],
        )

    execution: ExecutionResult | None = None
    function_results: list[FunctionResult] = []
    with DBConnection(connection_id) as dbc:
        for segment in case.segments:
            for attempt in range(case.metadata.retries + 1):
                execution = _execute_segment(dbc, segment.sql_parsed, case.metadata.timeout)
                if execution.status.success or attempt >= case.metadata.retries:
                    break
            if execution is None:
                raise RuntimeError("Execution never started")
            status = execution.status
            output = execution.output
            exit_on_failure = segment.directive.kwargs.get("exit_on_failure", True)
            func = registry.resolve(segment.directive.name)
            kwargs = {
                key: value
                for key, value in segment.directive.kwargs.items()
                if key != "exit_on_failure"
            }
            with execution_context(segment.sql_parsed, status, output):
                result = func(*segment.directive.args, **kwargs)
            function_results.append(result)
            if exit_on_failure and not result.success:
                break
    if execution is None:
        raise RuntimeError("Execution never started")
    return TestResult(
        case=case,
        status=execution.status,
        output=execution.output,
        function_results=function_results,
    )


def _format_query_output(rows: list[list[object]], columns: list[str] | None) -> str:
    """Format query results as newline-separated values for stdout."""
    if not columns:
        return ""

    parts = columns[:]
    for row in rows:
        parts.extend(str(val) if val is not None else "NULL" for val in row)

    return "\n".join(parts)


def _execute_segment(dbc: DBConnection, sql_parsed: SQLParsed, timeout: float | None = None) -> ExecutionResult:
    """Execute SQL using DBConnection and convert to ExecutionResult."""
    start = time.perf_counter()
    stdout = ""
    stderr = ""
    rows: list[list[object]] = []
    returncode = 0
    success = True

    try:
        # Execute each statement individually (like old SQLAlchemyConnector)
        statements = sql_parsed.statements
        if not statements:
            statements = []

        for statement in statements or []:
            result = dbc.query(statement.text, fetch="auto", include_columns=True)
            if result is not None:
                result_rows, columns = result
                rows = [list(row) for row in result_rows]
                stdout = _format_query_output(rows, columns)

        # Handle case where no statements but source exists
        if not statements and sql_parsed.source.strip():
            result = dbc.query(sql_parsed.source, fetch="auto", include_columns=True)
            if result is not None:
                result_rows, columns = result
                rows = [list(row) for row in result_rows]
                stdout = _format_query_output(rows, columns)

    except DBConnectionError as exc:
        success = False
        returncode = 1
        stderr = str(exc)
    except Exception as exc:
        success = False
        returncode = 1
        stderr = str(exc)

    duration = time.perf_counter() - start
    status = ExecutionStatus(success=success, returncode=returncode, duration_s=duration)
    output = ExecutionOutput(stdout=stdout, stderr=stderr, rows=rows)
    return ExecutionResult(status=status, output=output)


def run_cases(
    cases: Iterable[TestCase],
    connection_id: str,
    registry: FunctionRegistry,
    workers: int,
) -> list[TestResult]:
    parallel_cases = [case for case in cases if not case.metadata.serial]
    serial_cases = [case for case in cases if case.metadata.serial]
    results: list[TestResult] = []

    with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
        future_map = {
            executor.submit(run_test_case, case, connection_id, registry): case
            for case in parallel_cases
        }
        for future in concurrent.futures.as_completed(future_map):
            results.append(future.result())

    for case in serial_cases:
        results.append(run_test_case(case, connection_id, registry))

    return results



from __future__ import annotations

from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from sqlcheck.models import TestResult


def print_results(results: list[TestResult], engine: str | None = None) -> None:
    total = len(results)
    failures = [result for result in results if not result.success]
    passed = total - len(failures)
    console = Console()

    header = "SQLCheck"
    if engine:
        header += f" ({engine})"
    header += f" — {total} tests, {passed} passed"
    if failures:
        header += f", {len(failures)} failed"

    if failures:
        console.print("[bold]Failures:[/bold]")
        for result in failures:
            console.print(
                f"[red]FAIL[/red] {result.case.metadata.name}  [dim]{result.case.path}[/dim]"
            )
            for func_result in result.function_results:
                if not func_result.success:
                    message = func_result.message or "Expectation failed"
                    console.print(f"  {message}")
            if result.output.stderr:
                console.print(
                    Panel(
                        result.output.stderr.strip(),
                        title="STDERR",
                        border_style="red",
                    )
                )
            if result.output.stdout:
                console.print(
                    Panel(
                        result.output.stdout.strip(),
                        title="STDOUT",
                        border_style="yellow",
                    )
                )
        console.print()

    table = Table(box=box.ASCII, show_header=True, header_style="bold")
    table.add_column("STATUS", style="bold")
    table.add_column("TEST")
    table.add_column("DURATION", justify="right")
    table.add_column("PATH")

    for result in results:
        duration = f"{result.status.duration_s:.2f}s"
        status = "PASS" if result.success else "FAIL"
        status_style = "green" if result.success else "red"
        table.add_row(
            f"[{status_style}]{status}[/{status_style}]",
            result.case.metadata.name,
            duration,
            str(result.case.path),
        )

    console.print(table)
    console.print()
    console.print(f"[bold]{header}[/bold]")


__all__ = ["print_results"]



class DBConnectionError(RuntimeError):
    def __init__(
        self,
        message: str,
        *,
        conn_id: str,
        sql: str | None = None,
        original: Exception | None = None,
    ):
        super().__init__(message)
        self.conn_id = conn_id
        self.sql = sql
        self.original = original

    def __str__(self) -> str:
        # The main error message is already set in __init__, just add SQL context if helpful
        msg = super().__str__()

        # Optionally show a brief SQL snippet for context
        if self.sql and len(self.sql) > 200:
            # Only show snippet if SQL is long (short SQL is fine to show in full)
            sql_lines = self.sql.strip().split('\n')
            first_line = sql_lines[0] if sql_lines else self.sql
            sql_snippet = first_line[:100] + "..." if len(first_line) > 100 else first_line
            msg += f"\n\nSQL operation: {sql_snippet}"

        return msg



from __future__ import annotations

import time
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any
from urllib.parse import parse_qs, urlparse

from sqlalchemy import text
from sqlalchemy.engine import Connection, Engine, Result
from sqlalchemy.exc import SQLAlchemyError

from .engine import get_engine, dispose_all_engines
from .exceptions import DBConnectionError
from .models import ExecMeta, FetchMode
from .resolver import resolve_connection_url


@dataclass
class DBConnection:
    """
    SQLAlchemy 2.x helper with:
      - conn_id-driven URL resolution (env var lookup OR YAML file) OR direct URL usage
      - cached engines (Option A)
      - context-managed connection + transaction (Mode 1)
      - one query method that can return rows, rows+cols, dicts, etc.

    Example:
        db = DBConnection("snowflake_prod")
        with db as dbc:
            dbc.query("CREATE TEMP TABLE t1 AS SELECT 1")
            rows = dbc.query("SELECT * FROM t1")
            rows, cols = dbc.query("SELECT * FROM t1", include_columns=True)
            dicts = dbc.query("SELECT * FROM t1", as_dict=True)

    Connection resolution order (first match wins):
        1. Use default connection if conn_id is None (SQLCOMPARE_CONN_DEFAULT, DTK_CONN_DEFAULT)
        2. Direct URL (if contains "://")
        3. Environment variables: SQLCOMPARE_CONN_<NAME>, DTK_CONN_<NAME>
        4. YAML files: ~/.sqlcompare/connections.yml, ~/.dtk/connections.yml
    """

    conn_id: str | None = None
    engine_kwargs: dict[str, Any] = field(default_factory=dict)

    # runtime fields
    _engine: Engine | None = field(init=False, default=None)
    _conn: Connection | None = field(init=False, default=None)
    _tx: Any = field(init=False, default=None)

    # last execution metadata (handy for logging/debug)
    last_sql: str | None = field(init=False, default=None)
    last_elapsed_ms: int | None = field(init=False, default=None)
    last_rowcount: int | None = field(init=False, default=None)

    def __enter__(self) -> "DBConnection":
        url = resolve_connection_url(self.conn_id)

        # Check if this is a file-based connection
        file_path = None
        if "_file_path=" in url:
            parsed = urlparse(url)
            params = parse_qs(parsed.query)
            if "_file_path" in params:
                file_path = Path(params["_file_path"][0])
                # Remove the _file_path parameter from the URL
                url = url.split("?")[0]

        self._engine = get_engine(url, engine_kwargs=self.engine_kwargs)
        self._conn = self._engine.connect()
        # Mode 1: transaction per context
        self._tx = self._conn.begin()

        # If this is a file-based connection, load the file into DuckDB
        if file_path:
            self._load_file_into_duckdb(file_path)

        return self

    def __exit__(self, exc_type, _exc_val, _exc_tb) -> None:
        # Commit / rollback transaction
        try:
            if self._tx is not None:
                if exc_type is None:
                    self._tx.commit()
                else:
                    self._tx.rollback()
        finally:
            # Always close the SQLAlchemy Connection (returns DBAPI connection to pool,
            # OR fully closes it for NullPool / no-pool cases).
            if self._conn is not None:
                try:
                    self._conn.close()
                except Exception:
                    pass
            self._conn = None
            self._tx = None

    def _load_file_into_duckdb(self, file_path: Path) -> None:
        """
        Load a CSV or Excel file into DuckDB as a table.
        The table name will be the file stem (filename without extension).
        """
        self.create_table_from_file(file_path.stem, file_path)

    def _ensure_duckdb(self) -> None:
        if self._engine is None:
            raise DBConnectionError(
                "Connection is not open. Use `with DBConnection(...) as dbc:`",
                conn_id=self.conn_id or "<direct>",
            )
        if self._engine.dialect.name != "duckdb":
            raise DBConnectionError(
                "File loading requires a DuckDB connection.",
                conn_id=self.conn_id or "<direct>",
            )

    def _infer_column_type(self, table: str, column: str) -> str:
        for t in ["BIGINT", "DOUBLE", "TIMESTAMP", "DATE"]:
            q = (
                f"SELECT COUNT(*) FROM {table} "
                f'WHERE TRY_CAST("{column}" AS {t}) IS NULL AND "{column}" IS NOT NULL'
            )
            if self.conn.execute(text(q)).fetchone()[0] == 0:
                return t
        return "VARCHAR"

    def _ensure_excel_extension(self) -> None:
        try:
            self.conn.execute(text("LOAD excel"))
        except Exception:
            try:
                self.conn.execute(text("INSTALL excel"))
                self.conn.execute(text("LOAD excel"))
            except Exception as exc:
                raise RuntimeError(
                    "DuckDB excel extension is required to read XLSX files."
                ) from exc

    def create_table_from_file(self, table_name: str, file_path: str | Path) -> None:
        path = Path(file_path)
        suffix = path.suffix.lower()
        self._ensure_duckdb()

        if suffix == ".csv":
            read_query = f"SELECT * FROM read_csv_auto('{path}')"
        elif suffix == ".xlsx":
            self._ensure_excel_extension()
            read_query = f"SELECT * FROM read_xlsx('{path}')"
        else:
            raise ValueError(f"Unsupported file type: {suffix}")

        tmp = f"tmp_{uuid.uuid4().hex}"
        self.conn.execute(text(f"CREATE TABLE {tmp} AS {read_query}"))
        cols = [
            row[1]
            for row in self.conn.execute(text(f"PRAGMA table_info('{tmp}')")).fetchall()
        ]
        exprs = []
        for c in cols:
            t = self._infer_column_type(tmp, c)
            if t == "VARCHAR":
                exprs.append(f'"{c}"')
            else:
                exprs.append(f'CAST("{c}" AS {t}) AS "{c}"')
        self.conn.execute(
            text(f"CREATE TABLE {table_name} AS SELECT {', '.join(exprs)} FROM {tmp}")
        )
        self.conn.execute(text(f"DROP TABLE {tmp}"))

    @property
    def conn(self) -> Connection:
        if self._conn is None:
            raise DBConnectionError(
                "Connection is not open. Use `with DBConnection(...) as dbc:`",
                conn_id=self.conn_id,
            )
        return self._conn

    @property
    def raw_connection(self) -> Any:
        """Return the underlying DBAPI connection (e.g., DuckDB connection)."""
        if self._conn is None:
            raise DBConnectionError(
                "Connection is not open. Use `with DBConnection(...) as dbc:`",
                conn_id=self.conn_id,
            )
        return self._conn.connection.dbapi_connection

    def _run(self, sql: str, params: dict[str, Any] | None = None) -> Result:
        self.last_sql = sql
        t0 = time.perf_counter()
        try:
            # Using SQLAlchemy text() as requested
            res = self.conn.execute(text(sql), params or {})
            return res
        except SQLAlchemyError as e:
            # Extract just the database error message without the full SQL statement
            error_msg = str(e)
            # SQLAlchemy often includes the SQL in square brackets at the end
            # Format: "error message [SQL: long query here]"
            if "[SQL:" in error_msg:
                # Extract just the part before [SQL:
                db_error = error_msg.split("[SQL:")[0].strip()
            else:
                db_error = error_msg

            raise DBConnectionError(
                db_error,
                conn_id=self.conn_id,
                sql=sql,
                original=e,
            ) from None  # Suppress the original exception chain to avoid showing SQL twice
        finally:
            self.last_elapsed_ms = int((time.perf_counter() - t0) * 1000)

    def execute(self, sql: str, params: dict[str, Any] | None = None) -> ExecMeta:
        """
        Execute a statement and return lightweight metadata.
        Intended for DDL/DML (no row fetching).
        """
        res = self._run(sql, params=params)

        rowcount = res.rowcount if res.rowcount is not None else None
        cols = list(res.keys()) if res.returns_rows else None

        self.last_rowcount = rowcount
        return ExecMeta(
            elapsed_ms=self.last_elapsed_ms or 0, rowcount=rowcount, columns=cols
        )

    def get_table_columns(self, table_name: str) -> list[str]:
        """
        Get the actual column names from a table, using database-specific metadata queries
        to ensure correct case sensitivity (especially important for Snowflake).

        Args:
            table_name: Fully qualified table name (e.g., "schema.table" or "db.schema.table")

        Returns:
            List of column names with their actual case as stored in the database
        """
        # Detect if we're using Snowflake by checking the dialect
        try:
            dialect_name = self._engine.dialect.name.lower() if self._engine else None
        except Exception:
            dialect_name = None

        # For Snowflake, use DESCRIBE TABLE to get actual column names
        if dialect_name == "snowflake":
            try:
                # DESCRIBE TABLE returns columns with their actual case
                result = self.query(f"DESCRIBE TABLE {table_name}")
                # First column is the column name
                return [row[0] for row in result]
            except Exception:
                pass

        # Fallback: use SELECT * WHERE 1=0 and get column names from result
        _, columns = self.query(f"SELECT * FROM {table_name} WHERE 1=0", include_columns=True)
        return columns

    def query(
        self,
        sql: str,
        params: dict[str, Any] | None = None,
        *,
        fetch: FetchMode = "auto",
        include_columns: bool = False,
        as_dict: bool = False,
    ):
        """
        Execute SQL and (optionally) fetch results.

        Returns (depending on options + SQL):
          - None                           (DDL/DML with fetch="auto" or fetch=False)
          - list[tuple]                    (default for SELECT)
          - (list[tuple], list[str])       (include_columns=True)
          - list[dict[str, Any]]           (as_dict=True)
          - (list[dict[str, Any]], cols)   (as_dict=True + include_columns=True)
        """
        res = self._run(sql, params=params)

        if fetch == "auto":
            should_fetch = bool(res.returns_rows)
        else:
            should_fetch = bool(fetch)

        self.last_rowcount = res.rowcount if res.rowcount is not None else None

        if not should_fetch:
            return None

        columns = list(res.keys())

        if as_dict:
            mappings = res.mappings().all()  # list[RowMapping]
            dicts = [dict(m) for m in mappings]
            return (dicts, columns) if include_columns else dicts

        rows = res.all()  # list[Row]
        tuples = [tuple(r) for r in rows]
        return (tuples, columns) if include_columns else tuples

    @staticmethod
    def dispose_all_engines() -> None:
        """
        Force-close pooled connections and clear the engine cache.
        Useful in tests.
        """
        dispose_all_engines()

    @staticmethod
    def get_engine(url: str, *, engine_kwargs: dict[str, Any] | None = None) -> Engine:
        """
        Return a cached Engine for this URL (thread-safe).

        Default behavior:
        - Reuse engines (and their pools) across DBConnection instances for the same URL.
        - For SQLite in-memory URLs, auto-inject NullPool unless user overrides poolclass,
            so each `with DBConnection(...):` gets a truly fresh in-memory database.
        """
        return get_engine(url, engine_kwargs=engine_kwargs)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from jinja2 import Environment, StrictUndefined, TemplateSyntaxError, UndefinedError, Undefined


class TemplateRenderError(ValueError):
    """Raised when Jinja template rendering fails."""
    pass


class DirectiveUndefined(Undefined):
    """
    Custom undefined handler that treats undefined names as directive functions.

    This allows custom directives (not in the global registry) to still be parsed.
    Variables that are accessed but not called as functions will still raise errors.
    """
    def __call__(self, *args, **kwargs):
        # When an undefined name is called as a function, treat it as a directive
        return DirectiveFunction(self._undefined_name)(*args, **kwargs)

    def __str__(self):
        # When undefined variable is rendered as string, raise error
        self._fail_with_undefined_error()

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(name)
        # Return another undefined for chained access
        return self

    def __int__(self):
        self._fail_with_undefined_error()

    def __float__(self):
        self._fail_with_undefined_error()

    def __bool__(self):
        self._fail_with_undefined_error()


@dataclass(frozen=True)
class DirectiveMarker:
    """Marker object returned by directive functions during template parsing."""
    name: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    raw_text: str
    placeholder: str


# Global list to collect markers during rendering
_marker_collector: list[DirectiveMarker] = []


class DirectiveFunction:
    """Jinja-compatible directive function that collects markers."""

    def __init__(self, name: str):
        self.name = name

    def __call__(self, *args, **kwargs) -> str:
        """Called by Jinja during template rendering."""
        # Create unique placeholder
        marker_id = f"__DIRECTIVE_{len(_marker_collector)}__"

        # Reconstruct the call syntax
        args_str = ", ".join(repr(arg) for arg in args)
        kwargs_str = ", ".join(f"{k}={v!r}" for k, v in kwargs.items())
        all_args = ", ".join(filter(None, [args_str, kwargs_str]))
        raw_text = f"{{{{ {self.name}({all_args}) }}}}"

        # Create marker
        marker = DirectiveMarker(
            name=self.name,
            args=args,
            kwargs=kwargs,
            raw_text=raw_text,
            placeholder=marker_id
        )
        _marker_collector.append(marker)

        # Return placeholder that will be in rendered output
        return marker_id


def render_template(
    source: str,
    variables: dict[str, str] | None = None
) -> tuple[str, list[DirectiveMarker]]:
    """
    Render SQL source with Jinja2, extracting directives and variables.

    Uses {{ }} syntax for both variables and directive functions.

    Args:
        source: Raw SQL file content with Jinja template syntax
        variables: Dictionary of variable name -> value mappings

    Returns:
        Tuple of (rendered_source_with_placeholders, directive_markers)

    Raises:
        TemplateRenderError: If rendering fails
    """
    global _marker_collector
    _marker_collector = []  # Reset collector

    env = Environment(
        undefined=DirectiveUndefined,
        autoescape=False,
        finalize=lambda x: '' if x is None else x
    )

    # Register directive functions as Jinja globals
    env.globals['success'] = DirectiveFunction('success')
    env.globals['fail'] = DirectiveFunction('fail')
    env.globals['assess'] = DirectiveFunction('assess')

    template_vars = variables or {}

    try:
        template = env.from_string(source)
        rendered = template.render(**template_vars)
    except UndefinedError as exc:
        raise TemplateRenderError(f"Undefined variable in template: {exc}") from exc
    except TemplateSyntaxError as exc:
        raise TemplateRenderError(f"Template syntax error: {exc}") from exc
    except Exception as exc:
        raise TemplateRenderError(f"Failed to render template: {exc}") from exc

    # Collect markers and clear global list
    markers = list(_marker_collector)
    _marker_collector = []

    return rendered, markers



from __future__ import annotations

import sys

from sqlcheck.cli.app import app


def main(argv: list[str] | None = None) -> None:
    app(args=argv)


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import json
from dataclasses import asdict
from pathlib import Path
from typing import Any
from xml.etree import ElementTree

from sqlcheck.models import TestCase, TestResult


def build_plan_payload(case: TestCase) -> dict[str, Any]:
    return {
        "path": str(case.path),
        "name": case.metadata.name,
        "tags": case.metadata.tags,
        "serial": case.metadata.serial,
        "timeout": case.metadata.timeout,
        "retries": case.metadata.retries,
        "statements": [
            {"index": stmt.index, "text": stmt.text, "start": stmt.start, "end": stmt.end}
            for stmt in case.sql_parsed.statements
        ],
        "directives": [
            {"name": directive.name, "args": directive.args, "kwargs": directive.kwargs}
            for directive in case.directives
        ],
    }


def write_plan(result: TestResult, path: Path) -> None:
    payload = build_plan_payload(result.case)
    path.write_text(json.dumps(payload, indent=2), encoding="utf-8")


def write_case_plan(case: TestCase, path: Path) -> None:
    payload = build_plan_payload(case)
    path.write_text(json.dumps(payload, indent=2), encoding="utf-8")


def write_json(results: list[TestResult], path: Path) -> None:
    payload = []
    for result in results:
        payload.append(
            {
                "path": str(result.case.path),
                "name": result.case.metadata.name,
                "tags": result.case.metadata.tags,
                "serial": result.case.metadata.serial,
                "timeout": result.case.metadata.timeout,
                "retries": result.case.metadata.retries,
                "status": asdict(result.status),
                "output": asdict(result.output),
                "function_results": [asdict(item) for item in result.function_results],
                "success": result.success,
                "statements": [
                    {"index": stmt.index, "text": stmt.text, "start": stmt.start, "end": stmt.end}
                    for stmt in result.case.sql_parsed.statements
                ],
            }
        )
    path.write_text(json.dumps(payload, indent=2), encoding="utf-8")


def write_junit(results: list[TestResult], path: Path) -> None:
    testsuite = ElementTree.Element("testsuite", name="sqlcheck")
    testsuite.set("tests", str(len(results)))
    testsuite.set("failures", str(sum(1 for result in results if not result.success)))

    for result in results:
        testcase = ElementTree.SubElement(
            testsuite,
            "testcase",
            name=result.case.metadata.name,
            classname=str(result.case.path),
            time=f"{result.status.duration_s:.3f}",
        )
        if not result.success:
            failure = ElementTree.SubElement(testcase, "failure")
            messages = [
                item.message
                for item in result.function_results
                if not item.success and item.message
            ]
            detail = "\n".join(messages)
            failure.text = detail

    tree = ElementTree.ElementTree(testsuite)
    tree.write(path, encoding="utf-8", xml_declaration=True)
